/*
 * main.c
 *
 * Created: 30.03.2013 17:37:23
 *  Author: Lupin
 *
 * ATTiny25 clocked with 16 MHz (PLL clock)
 * Startup time should be short, so that bypass transistors will be turned on before ZXLD1350 starts operation
 */ 

#define F_CPU 16000000

// AVR includes
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/atomic.h>

/*
 Macros for peter danneggers senskey function
*/
struct bits {                           // Macro to access bits like variables
	uint8_t b0:1;
	uint8_t b1:1;
	uint8_t b2:1;
	uint8_t b3:1;
	uint8_t b4:1;
	uint8_t b5:1;
	uint8_t b6:1;
	uint8_t b7:1;
} __attribute__((__packed__));
#define SBIT(port,pin) ((*(volatile struct bits*)&port).b##pin)


/****** Port definitions *******/
#define LED_PORT		PORTB
#define RED_PIN			(1<<PB0)
#define GREEN_PIN		(1<<PB4)
#define BLUE_PIN		(1<<PB1)

// senskey stuff by peter Dannegger
#define SENSKEY_A0      SBIT( PORTB, 3 )        // connect to senspad
#define SENSKEY_A0_DDR  SBIT( DDRB,  3 )

#define SENSKEY_B0      SBIT( PORTB, 2 )        // connect to cap to A0
#define SENSKEY_B0_DDR  SBIT( DDRB,  2 )
#define SENSKEY_B0_PIN  SBIT( PINB,  2 )

// Threshold to detect if object was moved close or removed from touchpad
// This sets the touchpad sensitivity
#define TP_THRESHOLD	8




/*
 *  LED control macros
 */

#define RED_ON			LED_PORT &= ~RED_PIN;
#define RED_OFF			LED_PORT |= RED_PIN;

#define GREEN_ON		LED_PORT &= ~GREEN_PIN;
#define GREEN_OFF		LED_PORT |= GREEN_PIN;

#define BLUE_ON			LED_PORT &= ~BLUE_PIN;
#define BLUE_OFF		LED_PORT |= BLUE_PIN;

#define LEDS_OFF		LED_PORT |= RED_PIN | GREEN_PIN | BLUE_PIN;

// Touchpad macro definitions (main loop)
#define TP_UP			0
#define TP_DOWN			1
#define TP_HOLD			2

// HSV color space stuff
int32_t h;
#define HUE_FBITS		9 // How many fraction bits for hue
#define HUE_FMAX		((uint32_t)(((uint32_t)1<<HUE_FBITS)-1))
#define HUE_MAX			(((int32_t)360)<<HUE_FBITS)

// Converted HSV->RGB values
uint8_t r, g, b;

// Speed of LED fading
uint32_t speed;

// To turn LED PWM on/off
uint8_t led_on;

// Counter for mainloop to read a new touchpad value (at 5Hz)
// Will be reset by the main loop
volatile uint16_t smplctr;




/*
senskey function by peter dannegger:
http://www.mikrocontroller.net/topic/156809#1483874

A0      B0      B0_PIN
0       0               full discharge
loop:
Z       Z               gap
Z       1               partial charge over pin A0 and senspad
Z       Z               gab
0       Z               partial charge over pin B0 only
0       Z       1       end
*/
#define MAX_CYCLE       2000
uint16_t senskey( void )
{
	uint16_t i = MAX_CYCLE + 1;

	SENSKEY_A0 = 0;
	SENSKEY_A0_DDR = 1;
	SENSKEY_B0 = 0;
	SENSKEY_B0_DDR = 1;           // discharge cap

	ATOMIC_BLOCK(ATOMIC_FORCEON){
		do{
			SENSKEY_A0_DDR = 0;       // A0 = tristate
			SENSKEY_B0 = 1;           // B0 = weak high
			SENSKEY_B0_DDR = 1;       // B0 = strong high
			SENSKEY_B0_DDR = 0;       // B0 = weak high
			SENSKEY_B0 = 0;           // B0 = tristate
			SENSKEY_A0_DDR = 1;       // A0 = strong low
			if( --i == 0 )
			break;                          // timeout
		}while( SENSKEY_B0_PIN == 0 );      // until charged
	}
	SENSKEY_B0_DDR = 1;           // discharge cap
	return MAX_CYCLE - i;         // big value = key touched or timeout
}



/*
 * Converts HUE of HSV colors space into RGB color space
 * Saturation and Value component of HSV color space are always "1"
 * This function only moves along the outer circle of the HSV cone
 */
void hsv_to_rgb(void)	
{
  uint8_t f;
  uint32_t ht;

  ht = h / 60;
  f = (uint8_t)(((ht&HUE_FMAX) * 255)>>HUE_FBITS);

  switch(ht>>HUE_FBITS) {
	case 0:
		r = 255; g = f; b = 0; break;
	case 1:
		r = 255 - f; g = 255; b = 0; break;
	case 2:
		r = 0; g = 255; b = f; break;
	case 3:
		r = 0; g = 255 - f; b = 255; break;			
	case 4:
		r = f; g = 0; b = 255; break;
	case 5:
		r = 255; g = 0; b = 255 - f; break;

	default:
		r = 0; g = 0; b = 0; break;
  }
}

/*
 * Soft PWM done by timer1
 * Also serves as a timebase for the mainloop, updates the hue and
 * calculates new RGB values when the PWM period is finished
 */
SIGNAL(TIMER1_OVF_vect) {
    static uint8_t counter;

	// Counter for main loop
	smplctr++;

	// Check if the lamp should be on
	if(led_on==1) {
		if (counter == 0) {
			if(r>0) RED_ON else RED_OFF;
			if(g>0) GREEN_ON else GREEN_OFF;
			if(b>0) BLUE_ON else BLUE_OFF;
		}
	
		if (counter != 255) {
			if (counter == r) RED_OFF;
			if (counter == g) GREEN_OFF;
			if (counter == b) BLUE_OFF;
		} else {
			// 255 is reached, calculate new RGB values
			// This should actually not be done inside an IRQ because
			// the calculation can take some time and IRQ routines should be short
			// But apparently this works very well...
			h+=speed;
			if(h>=HUE_MAX) h-=HUE_MAX;
			hsv_to_rgb();
		}
		counter++;
	} else {
		LEDS_OFF;
	}
}

/*
 * Simple moving average filter
 */
// Filter order (must be 2^n)
#define FILORD		8	// This is actually the length of the filter... order is FILORD-1
#define FILSHIFT	3   // log2(FILORD) - C macro for this...?!

// Moving average lowpass filter
int16_t lpfilter(int16_t sample) {
	static int16_t z[FILORD];
	static int8_t ctr;
	int8_t i;
	int16_t fil;
	
	z[ctr&(FILORD-1)] = sample;
	ctr++; fil=0;
	for(i=0 ; i<FILORD ; i++) {
		fil += z[i];
	}
	return fil>>FILSHIFT;
}

// Settle the lowpass filter to some sampled value
void lpfilter_settle(int16_t sample) {
	int8_t i;
	for(i=0 ; i<FILORD ; i++) lpfilter(sample);
}


int	main(void)
{
	// Filtered and current touchpad value
	int16_t tp_filter, tp_current;
	int8_t tp_state;
	
	// Set LED port to output and turn all bypass transistors on (LEDs off)
	PORTB = RED_PIN | GREEN_PIN | BLUE_PIN;
	DDRB = RED_PIN | GREEN_PIN | BLUE_PIN;

	// Timer1 is clocked with CK/2 -> 8MHz
	TCCR1 = (1<<CS11);
	// Enable Timer1 overflow interrupt
	// Timer1 overflows at 8MHz/256 -> 31250 Hz
	// LED PWM is 8 Bit, so the PWM-Frequency of the LEDs is 31250Hz/256 -> 122Hz
	// The flickering should not be noticeable...
	TIMSK = (1<<TOIE1);
	
	// Initialize some variables
	h=speed=led_on=0;
	// Assume the touch area is not touched at the moment
	tp_state = TP_UP;
	
	// Initialize the lowpass filter
	tp_filter = tp_current = senskey();
	lpfilter_settle(tp_current);
	
	// Activate timer IRQ
	sei();
	
	// Main loop
	while(1) {		

		// Read touchpad and filter signals at 5Hz
		if(smplctr >= 6250) {
			
			// Read touchpad value
			tp_current = senskey();
			
			// Sense a touch event by comparing the current value
			// to the lowpass filtered touchpad value
			
			// Check if the touchpad is pressed
			// If the touchpad already was pressed before then this event can only be triggered by
			// applying more force to the sensor area or directly touching the sensor area (do nothing if this happens)
			if(tp_current < (tp_filter - TP_THRESHOLD) && tp_state == TP_UP) {
				// "Button down event"
				
				// New touchpad state
				tp_state = TP_DOWN;
				
				// Check if the lamp is on
				if(led_on == 1) {
					// It is on - Turn off the lamp
					led_on = 0;
				} else {
					// It is off - Turn on the lamp
					led_on = 1;
					speed = 1;
				}
				
				// Settle the lowpass filter to the current touchpad value
				// so that a touchpad change will be sensed relative to this value
				lpfilter_settle(tp_current);
			
			
			} else if(tp_current > (tp_filter + TP_THRESHOLD) && tp_state != TP_UP) {
				// "Button up event"
				
				// New touchpad state
				tp_state = TP_UP;
				
				// Settle the lowpass filter to the current touchpad value
				// so that a touchpad change will be sensed relative to this value
				lpfilter_settle(tp_current);
			
			
			} else if(tp_state != TP_UP) {
				// "Button hold event"
				
				// New touchpad state
				tp_state = TP_HOLD;

				// Fading is inactive (speed=0) - Choose the animation speed
				speed+=1;

			}
			
			// Filter touchpad value
			tp_filter = lpfilter(tp_current);
			
			// Start counting the timer from 0
			smplctr = 0;
		}
	
	};
	
	return 0;
}
